// ##########################################################################
// ####                                                                  ####
// ####      WarpEd - A portable text editor for OS/2 Warp and Linux     ####
// ####    ===========================================================   ####
// ####                                                                  ####
// #### console.h                                                        ####
// ####                                                                  ####
// #### Version 0.80  --  October 06, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 1996  Thomas Dreibholz                             ####
// ####               1997  Molbachweg 7                                 ####
// ####                     51674 Wiehl/Germany                          ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CONSOLE_H
#define CONSOLE_H

#include "terminal.h"

#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <term.h>
#include <bool.h>
#include <ncurses.h>
#undef clear


// #######################################################################
// ===== Farbverwaltung ==================================================
// #######################################################################

struct ColorPrefs
{
   int FrontPen;
   int BackPen;
   int Style;
   int BlockFrontPen;
   int BlockBackPen;
   int BlockStyle;
   int InfoFrontPen;
   int InfoBackPen;
   int InfoStyle;
};

#define NormalColorPair 1
#define BlockColorPair  2
#define InfoColorPair   3

#define DefaultFrontPen      3
#define DefaultBackPen       0
#define DefaultStyle         0
#define DefaultBlockFrontPen 1
#define DefaultBlockBackPen  7
#define DefaultBlockStyle    0
#define DefaultInfoFrontPen  6
#define DefaultInfoBackPen   0
#define DefaultInfoStyle     0


// #######################################################################
// ===== Tastendefinitionen ==============================================
// #######################################################################

#define BlockMode 10000
#define ControlSeq(x) (x - 64)

#define KeyF1                 KEY_F(1)
#define KeyF2                 KEY_F(2)
#define KeyF3                 KEY_F(3)
#define KeyF4                 KEY_F(4)
#define KeyF5                 KEY_F(5)
#define KeyF6                 KEY_F(6)
#define KeyF7                 KEY_F(7)
#define KeyF8                 KEY_F(8)
#define KeyF9                 KEY_F(9)
#define KeyF10                KEY_F(10)
#define KeyF11                KEY_F(11)
#define KeyF12                KEY_F(12)
#define KeyInsert             KEY_IC
#define KeyHome               KEY_HOME
#define KeyPageUp             KEY_PPAGE
#define KeyDelete             KEY_DC
#define KeyEnd                KEY_END
#define KeyPageDown           KEY_NPAGE
#define KeyCursorUp           KEY_UP
#define KeyCursorLeft         KEY_LEFT
#define KeyCursorDown         KEY_DOWN
#define KeyCursorRight        KEY_RIGHT
#define KeyBlockInsert        BlockMode|KEY_IC
#define KeyBlockHome          BlockMode|KEY_HOME
#define KeyBlockPageUp        BlockMode|KEY_PPAGE
#define KeyBlockDelete        BlockMode|KEY_DC
#define KeyBlockEnd           BlockMode|KEY_END
#define KeyBlockPageDown      BlockMode|KEY_NPAGE
#define KeyBlockCursorUp      BlockMode|KEY_UP
#define KeyBlockCursorLeft    BlockMode|KEY_LEFT
#define KeyBlockCursorDown    BlockMode|KEY_DOWN
#define KeyBlockCursorRight   BlockMode|KEY_RIGHT
#define KeyBackspace          KEY_BACKSPACE
#define KeyShiftBackspace     ControlSeq('Y')
#define KeyReturn             0x0d
#define KeyEscape             0x1b
#define KeyTab                9


// #######################################################################
// ===== Terminal-Klasse =================================================
// #######################################################################


class OutputTerminalClass : public TerminalClass
{
   public:
   int        zWidth;
   int        zHeight;
   ColorPrefs zColorPrefs;
   int        zCursorX;
   int        zCursorY;
   int        zShowCursor;

   int  OpenTerminal(ColorPrefs *cprefs);
   void CloseTerminal();
   
   void Clear();
   void Bell();

   void BeginWriting();
   void EndWriting();
   void WriteLine(int x, int y, int width,
                  int offset, int m1, int m2, 
                  char       *buffer);
   void WriteString(int x, int y, int style, int frontPen, int backPen,
                    char *str);
   
   void ShowSysCursor(int x, int y);
   void HideSysCursor();
   
   int  WaitForKey(int secs, int usecs);    

   private:
   int  zWriteLevel;
};

#endif
