// ##########################################################################
// ####                                                                  ####
// ####      WarpEd - A portable text editor for OS/2 Warp and Linux     ####
// ####    ===========================================================   ####
// ####                                                                  ####
// #### ed.cc                                                            ####
// ####                                                                  ####
// #### Version 0.80  --  October 06, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 1996  Thomas Dreibholz                             ####
// ####               1997  Molbachweg 7                                 ####
// ####                     51674 Wiehl/Germany                          ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>
#include <stdlib.h>

#include "system.h"
#include "edit.h"
#include "edit.icc"

#include "const.h"        
#include "command.h"


ColorPrefs gWarpEdColors =
{
  DefaultFrontPen,       DefaultBackPen,       DefaultStyle,
  DefaultBlockFrontPen,  DefaultBlockBackPen,  DefaultBlockStyle,
  DefaultInfoFrontPen,   DefaultInfoBackPen,   DefaultInfoStyle
};
 
int gLogo        = 0;
int gControl     = 0;
struct WarpEd *gWarpEd;  


void LoadDefaultConfiguration()
{
   gWarpEd->FastScrollX     = 0;
   gWarpEd->FastScrollY     = 0;
   gWarpEd->OverwriteBlocks = 1;
   gWarpEd->AutoIndent      = 1;
   gWarpEd->AutoCursorJump  = 0;
   gWarpEd->AutoBackup      = 1;
   gWarpEd->FindText[0] = 0x00;
   gWarpEd->FileName[0] = 0x00;
   
   #if defined OS2_WARP
      strcpy(gWarpEd->TmpName,"warped.tmp");
   #endif
   #if defined LINUX
      strcpy(gWarpEd->TmpName,"warped.tmp");
   #endif
}

// Voreinstellungen laden
int LoadConfiguration()
{
   FILE *fh;
   int   i,l;
   char  config[cConfigLineLength]; 
   char  var[cConfigLineLength]; 

   LoadDefaultConfiguration();
   fh = fopen(gWarpEdConfigName,"r");
   if(fh != NULL) {
      l = 0;
      while(!(feof(fh))) {
         l++;
         i = fscanf(fh,"%s = %s",&config,&var); 
         if(!(strcmp(config,"TmpName"))) {
            strcpy(gWarpEd->TmpName,var);
         }
         else if(!(strcmp(config,"FastScrollX"))) {
            gWarpEd->FastScrollX = atoi(var);
         }
         else if(!(strcmp(config,"FastScrollY"))) {
            gWarpEd->FastScrollY = atoi(var);
         }
         else if(!(strcmp(config,"AutoIndent"))) {
            gWarpEd->AutoIndent = atoi(var);
         }
         else if(!(strcmp(config,"AutoBackup"))) {
            gWarpEd->AutoBackup = atoi(var);
         }
         else if(!(strcmp(config,"AutoCursorJump"))) {
            gWarpEd->AutoCursorJump = atoi(var);
         }
         else if(!(strcmp(config,"OverwriteBlocks"))) {
            gWarpEd->OverwriteBlocks = atoi(var);
         }
         else {
            printf("Configuration file <%s:%d> Syntax error!\n",gWarpEdConfigName,l);
            return(0);
         }
      }
      fclose(fh);
   }
   return(1);
}


/* WarpEd initialisieren */
void InitWarpEd()
{
   if(LoadConfiguration() == 0) {
      exit(1);
   }

   int ok = gWarpEd->Terminal.OpenTerminal(&gWarpEdColors);
   if(ok == FALSE) {
      puts("ERROR: Unable to initialize terminal!");
      exit(1);
   }                                            

   /* Initialisieren des Eds */
   gWarpEd->Ed.init(&gWarpEd->Terminal,0,0,
              gWarpEd->Terminal.zWidth,gWarpEd->Terminal.zHeight - 1,
              cRows,cCols,gWarpEd->TmpName);
   if(gWarpEd->Ed.Bad()) {
      puts("ERROR: Out of memory!");
      exit(0); 
   }
   gWarpEd->Ed.SetFastScroll(gWarpEd->FastScrollX,gWarpEd->FastScrollY);
   gWarpEd->Ed.SetAutoBackup(gWarpEd->AutoBackup);
   gWarpEd->Ed.SetAutoIndent(gWarpEd->AutoIndent);
   gWarpEd->Ed.SetAutoCursorJump(gWarpEd->AutoCursorJump);
   gWarpEd->Ed.SetOverwriteBlocks(gWarpEd->OverwriteBlocks);
   
   ShowLogo();
   UpdateInfo(NULL,0,cUpdateTimeAndCursor|cUpdateStatus);
}


/* WarpEd schlieen */
void CloseWarpEd()
{
   gWarpEd->Terminal.CloseTerminal();
}


// #######################################################################
// ===== Tastatur-Handler ================================================
// #######################################################################


/* Standard-Tastaturhandler */
void KeyboardHandler(int key)
{
   switch(key) {
      case KeyHome:
        gWarpEd->Ed.CursorHome();
       break;
      case KeyEnd:
        gWarpEd->Ed.CursorEnd();
       break;
      case KeyPageUp:
        gWarpEd->Ed.PageUp();
       break;
      case KeyPageDown:
        gWarpEd->Ed.PageDown();
       break;
      case KeyCursorLeft:
        gWarpEd->Ed.CursorLeft();
       break;
      case KeyCursorRight:
        gWarpEd->Ed.CursorRight();
       break;
      case KeyCursorUp:
        gWarpEd->Ed.CursorUp();
       break;
      case KeyCursorDown:
        gWarpEd->Ed.CursorDown();
       break; 
      case KeyDelete:
        gWarpEd->Ed.Del();
       break;
      case KeyF11:
        gWarpEd->Ed.BlockStart();
       break;
      case KeyF12:
        gWarpEd->Ed.BlockEnd();
       break;
      case KeyBlockHome:
        gWarpEd->Ed.BlockCursorHome();
       break;
      case KeyBlockEnd:
        gWarpEd->Ed.BlockCursorEnd();
       break;
      case KeyBlockPageUp:
        gWarpEd->Ed.BlockPageUp();
       break;
      case KeyBlockPageDown:
        gWarpEd->Ed.BlockPageDown();
       break;
      case KeyBlockCursorLeft:
        gWarpEd->Ed.BlockCursorLeft();
       break;
      case KeyBlockCursorRight:
        gWarpEd->Ed.BlockCursorRight();
       break;
      case KeyBlockCursorUp:
        gWarpEd->Ed.BlockCursorUp();
       break;
      case KeyBlockCursorDown:
        gWarpEd->Ed.BlockCursorDown();
       break; 
      case KeyBlockDelete:
        gWarpEd->Ed.ShiftDel();
       break;
      case KeyShiftBackspace:
        gWarpEd->Ed.ShiftBackspace();
       break;
      case KeyReturn:
        gWarpEd->Ed.Return();
       break;
      case KeyBackspace:
        gWarpEd->Ed.Backspace();
       break;
      default:
        if(key > 0) {
           if(key < 256) {
              gWarpEd->Ed.InsertChar((char)key);
           }
        }
       break;
   }
}
	

/* WarpEd ausfhren */
int ExecuteWarpEd(int key)
{
   int quit = FALSE;
   
   switch(key) {
      case KeyTab:
         quit = TRUE;
       break;
      case KeyEscape:
        quit = GetCommand("WarpEd> ");
       break;
      case KeyReturn:
        KeyboardHandler(key);
       break;
      default:
         // Control-Sequenzen abarbeiten
         for(char c = 'A';c <= 'Z';c++) {
           if(key == ControlSeq(c)) {
               if(gControl != 0x00) {
                  gControl = ControlSequence(gControl,key,&quit);
               }
               else {
                  gControl = ControlSequence(key,0,&quit);
               }
               return(quit);
            }
         }
         KeyboardHandler(key);
       break;
   }

   return(quit);
}


/* Hauptprogramm */
void main(int argc, char **argv)
{
   int key;
   int quit;

   gWarpEd = new WarpEd;
   if(gWarpEd == NULL) {
      puts("Out of memory!");
      exit(1);
   }


   // Ed initialisieren
   InitWarpEd();
   if(argc > 1) {
      Load(argv[1]);
   }

   // Hauptprogramm
   quit = FALSE;
   do {
      // Zeichen einlesen
      key = gWarpEd->Terminal.WaitForKey(0,950);
      if(key > 0) {
         if(gLogo == 1) {
            gWarpEd->Ed.UpdateDisplay();
            gLogo = 0;
         }

         // Eingabe verarbeiten
         quit = ExecuteWarpEd(key);

         // Ausgabe von Fehlermeldungen
         int stat = gWarpEd->Ed.Status();

         if(stat != 0) {
            if(stat > 8) {
               stat = 8;
            }
            UpdateInfo(NULL,stat,cUpdateTimeAndCursor|cUpdateStatus);
         }
         else {
            UpdateInfo(NULL,0,cUpdateTimeAndCursor);
         }
      }
      else {
        UpdateInfo(NULL,0,cUpdateTimeAndCursor);
      }
   } while(quit == FALSE);


   // Ed schlieen
   CloseWarpEd();
   delete gWarpEd;
}
