// ##########################################################################
// ####                                                                  ####
// ####      WarpEd - A portable text editor for OS/2 Warp and Linux     ####
// ####    ===========================================================   ####
// ####                                                                  ####
// #### edit.cc                                                          ####
// ####                                                                  ####
// #### Version 0.80  --  October 06, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 1996  Thomas Dreibholz                             ####
// ####               1997  Molbachweg 7                                 ####
// ####                     51674 Wiehl/Germany                          ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <bool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>

#include "str.h"
#include "terminal.h"
#include "row.h"
#include "row.icc"
#include "lines.h"
#include "lines.icc"
#include "edit.h"
#include "edit.icc"


#define min(a,b) ((a) <= (b) ? (a) : (b))
#define max(a,b) ((b) <= (a) ? (a) : (b))


// #######################################################################
// ===== Initialisierung =================================================
// #######################################################################


// Standard-Konstruktor
Editor::Editor()
{
   isGood = false;
}


/* Initialisieren des Editors */
void Editor::init(TerminalClass* term,
               int x, int y, int w, int h,
               int mr, int mc,
               char* tempName)
{
   lineMgr.init(mr,mc);
   edLine.init(term,x,y,w,mc,0,0);

   // Initialisieren der Grundparameter
   status   = ok;
   leftEdge = x; 
   topEdge  = y;
   width    = w;
   height   = h;
   cursorX  = 0;
   cursorY  = 0;
   offsetX  = 0;
   offsetY  = 0;
   changed  = false; 
   maxRows  = mr;
   maxCols  = mc;
   isGood   = false;
   Terminal = term;
     
   blockStartX = -1;
   blockStartY = -1;
   blockEndX   = -1;
   blockEndY   = -1;
   markX1 = -1;
   markX2 = -1;      
   markY1 = -1;      
   markY2 = -1;      
   blockStatus = noBlock;
   skipUpdate = 0;

   tmpName           = tempName;
   autoCursorJump    = false;
   overwriteBlocks   = true;
   autoIndent        = false;
   autoBackup        = true;
   readOnly          = false;
   fastScrollX       = 0;
   fastScrollY       = 0;
 
   if(lineMgr.Bad()) return;
   if(edLine.Bad()) return;
   edLine.ShowCursor();
   edLine.SysCursorUpdate(true);
   edLine.SetFastScroll(fastScrollX);
   isGood = true;
}


/* Editor freigeben */
Editor::~Editor()
{ }


/* Position und Gre des Editors ndern */
void Editor::Resize(int x, int y, int w, int h)
{
   leftEdge = x;
   topEdge  = y;
   width    = w;
   height   = h;
 
   // Editierzeile modifizieren
   edLine.Resize(leftEdge,topEdge,width);

   // Komplettes Update
   skipUpdate++;
   SetCursor(cursorY,cursorX);
   skipUpdate--;
   UpdateDisplay();
}


// #######################################################################                                                                    
// ===== Update des Displays und der Editierzeile ========================
// #######################################################################                                                                    


/* nderung der Y-Position des Cursors: 
   Cursor in neuer Zeile setzen und in alter entfernen */
void Editor::UpdateEditLine()
{
   // Editierzeile updaten
   char row[maxCols + 4];
   
   edLine.Resize(leftEdge,topEdge + (cursorY - offsetY),width);
   lineMgr.GetLine(cursorY,row);
   edLine.SetCursor(cursorX);
   edLine.SetOffset(offsetX);
   edLine.SetText(row);
   edLine.ShowCursor();
   
   // Markierung
   if(blockStatus == blockEnd) { 
      if((cursorY == markY1) && (cursorY == markY2)) {
         edLine.SetBlock(markX1,markX2);
      }
      else if(cursorY == markY1) {
         edLine.SetBlock(markX1,maxCols);
      }
      else if(cursorY == markY2) {
         edLine.SetBlock(0,markX2);
      }
      else if((cursorY > markY1) && (cursorY < markY2)) {
         edLine.SetBlock(0,maxCols);
      }
      else {
         edLine.SetBlock(-1,-1);
      }
   }
   else {
      edLine.SetBlock(-1,-1);
   }

   if(skipUpdate == 0) {
      edLine.Update();
   }
}                                          


/* Neuberechnung von offsetX und cursorX, gegebenenfalls Update
   des gesamten Displays */
void Editor::RethinkXPosition()
{
   offsetX = edLine.Offset();
   cursorX = edLine.Cursor();      

   if(skipUpdate == 0) {
      if(edLine.Changed() == true) {
         RefreshDisplay(-1,-1);
      } 
   }
}


/* Update mit Auslesen der Zeilen aus dem lineMgr */
void Editor::UpdateDisplay()
{
   skipUpdate++;
   UpdateEditLine(); 
   skipUpdate--;
   LoadCurrentLine();
   RefreshDisplay(-1,-1);
}


/* Komplettes Update des Displays: Ausgabe aller Line */
void Editor::RefreshDisplay(int start, int end)
{
   int i;
 
   if(skipUpdate == 0) {
      int   m1;
      int   m2;

      Terminal->BeginWriting();
      for(i = 0;i < height;i++) {
         if((start==-1) || (offsetY + i >= start) && (offsetY + i <= end)) {
 
            // Markierung einstellen
            if(blockStatus == blockEnd) {
               if((offsetY + i > markY1) && (offsetY + i < markY2)) {
                  m1 = 0; 
                  m2 = maxCols;
               }
               else if((offsetY + i == markY1) && (offsetY + i == markY2)) {
                  m1 = markX1;
                  m2 = markX2;
               }
               else if(offsetY + i == markY1) {
                  m1 = markX1;
                  m2 = maxCols;
               }
               else if(offsetY + i == markY2) {
                  m1 = 0;
                  m2 = markX2;
               }
               else {
                  m1 = -1;
                  m2 = -1;
               }
            }
            else {
               m1 = -1;
               m2 = -1;
            }
         
            // Zeile aus dem Speicher holen und ausgeben
           if(offsetY + i != cursorY) {
               char row[maxCols + 4];
               lineMgr.GetLine(offsetY + i,row);
               Terminal->WriteLine(leftEdge,topEdge + i,width,offsetX,
                                    m1,m2,row);
            }
            else {
               edLine.SetBlock(m1,m2);
               edLine.Update();
            }
         }
      }
      Terminal->EndWriting();
   }
}


// #######################################################################                                                                    
// ===== Einfgen und Lschen von Zeichen ================================
// #######################################################################                                                                    


/* Zeichen einfgen */
bool Editor::InsertChar(char chr)
{
   if(readOnly) return(false);

   // Block falls notwendig lschen
   OverwriteBlock();
   UnmarkBlock(true);

   // Zeichen einfgen, auf Fehler berprfen
   if(edLine.InsertChar(chr) == false) {
      status = lineTooLong;
      return(false);
   }
 
   // Editiereile in Zeilenspeicher kopieren
   SaveCurrentLine();

   // Markierung falls vorhanden lschen
   UnmarkBlock(true);

   // Neue X-Position seten
   RethinkXPosition();
   
   changed = true;
   return(true);
}


/* Zeichen lschen (Del-Taste) */
void Editor::Del()
{
   // Falls vorhanden, markierten Bereich lschen und Rckkehr
   if((readOnly) || (OverwriteBlock())) {
      return;  
   }

   // Markierung aufheben
   UnmarkBlock(true);

   char row1[maxCols + 4];
   edLine.GetText(row1);
   int l = strlen(row1);

   // Einfach Zeichen lschen
   if((cursorX < l) || (cursorY == maxRows-1)) {
      edLine.Del();
      SaveCurrentLine();
      changed = true;
   }

   // Sonst nachfolgende Zeile an aktuelle anfgen
   else {
      char row2[maxCols + 4];
      lineMgr.GetLine(cursorY + 1,row2);

      // Nur anfgen, wenn im Speicher noch genug Plat ist
      if(cursorX + (int)strlen(row2) < maxCols) {
         // Falls Zeile vorher beendet, Leereichen einfgen
         for(int i = l;i < cursorX;i++) { 
            Insert(' ',row1,i);
         }

         // Anfgen und zurckkopieren
         strcat(row1,row2);
         lineMgr.SetLine(cursorY,row1);
         edLine.SetText(row1);
         lineMgr.DeleteLine(cursorY + 1);

         // Update des Displays
         RefreshDisplay(cursorY,cursorY + height);
         changed = true;       
      }
      else {
         status = lineTooLong;
      }
   }
}


/* Zeile bis um Ende lschen */
void Editor::ShiftDel()
{
   if(readOnly) return;

   char row[maxCols + 4];
   edLine.GetText(row);

   // Steht der Cursor vor dem Zeilenende: Lschen bis um Zeilenende
   if((int)strlen(row) > cursorX) {
      row[cursorX] = 0x00;
      edLine.SetText(row);
      edLine.Update();
      SaveCurrentLine();
      changed = true;
   }
   // Sonst vorhergehende Zeile anfgen
   else {
      Del();
   }
}


/* Return-Taste */
void Editor::Return()
{
   if(readOnly) return;
  
   // Block falls notwendig lschen und Markierung aufheben
   OverwriteBlock();
   UnmarkBlock(true);
 
   if(cursorY < maxRows-1) {
      int  i,j;
      char row0[maxCols + 4];
      char row1[maxCols + 4];
      char row2[maxCols + 4];

      // Aktuelle Zeile aufteilen
      edLine.GetText(row0);
      Copy(row0,0,cursorX - 1,row1);
      Copy(row0,cursorX,strlen(row0) - 1,row2);
   
      // Zeile einfgen
      i = lineMgr.InsertLine(cursorY+1);     
      if(i == true) {
         // Zeilenteile einseten
         lineMgr.SetLine(cursorY,row1);
         lineMgr.SetLine(cursorY + 1,row2);
         
         int xpos = cursorX;
         int ox = offsetX;
         int oy = offsetY;

         skipUpdate++;

         // Neue Positionen berechnen
         cursorX = 0;
         offsetX = 0;
         cursorY++;
         if(cursorY >= offsetY + height - 1) {
            offsetY++;
         }
         UpdateEditLine();
         
         // Auto-Indent Funktion
         if((autoIndent) && (strlen(row2) == 0)) {
            if(cursorY > 0) {
               lineMgr.GetLine(cursorY - 1,row0);
               j = strlen(row0);

               if(j >= xpos) {
                  for(i = 0;i < j;i++) {
                     if(row0[i] != ' ') { 
                        break;
                     }
                  }
               }
               else {
                  i = xpos;
               }

               SetCursor(cursorY,i);
            }
         }
         skipUpdate--;

         // Update des vernderten Bereiches
         if((ox != offsetX) || (oy != offsetY)) {
            UpdateDisplay();
  
         }
         else {
            RefreshDisplay(cursorY - 1,cursorY + height);
         }
         
         changed = true;
      }
      else {
         status = bufferFull;
      }
   }
   else {
      status = bufferFull;
   }
}


/* Backspace-Taste */
void Editor::Backspace()
{
   // Evtl. Vorhandenen Block lschen
   if((readOnly) || (OverwriteBlock())) {
      return;
   }

   // Sonst Zeichen vor dem Cursor lschen
   else { 
      // Einelnes Zeichen lschen
      if(cursorX > 0) {
         edLine.Backspace();
         SaveCurrentLine();
         RethinkXPosition();
         changed = true;
      }
      // Text an vorherige Zeile anhngen
      else {
         if(cursorY > 0) {
            char row1[maxCols + 4];
            char row2[maxCols + 4];

            // Zeilen auslesen
            lineMgr.GetLine(cursorY - 1,row1);
            edLine.GetText(row2);
            int l1 = strlen(row1);
            int l2 = strlen(row2);

            if(l1 + l2 < maxCols) {
               strcat(row1,row2);
               lineMgr.SetLine(cursorY - 1,row1);
               lineMgr.DeleteLine(cursorY);
         
               int ox = offsetX;
               int oy = offsetY;

               // Neue Y-Position des Cursors berechnen
               cursorY--;
               if(cursorY == offsetY) {
                  offsetY--;
                  if(offsetY < 0) {
                     offsetY = 0;
                  }
               }
   
               // Neue X-Position des Cursors berechnen
               cursorX = l1;
               if(cursorX > width)
               offsetX = cursorX - (width / 3);
               if(offsetX < 0)
                  offsetX = 0;

               // Update des Displays
               if((offsetY != oy) || (offsetX != ox)) {
                  UpdateDisplay();
               }
               else {
                  UpdateEditLine();
                  RefreshDisplay(cursorY - 1,cursorY + height);
               }

               changed = true;   
            }
            else
               status = lineTooLong;
         }
         else
            status = topOfText;
      }
   }
}


/* Zeile bis zum Anfang lschen */
void Editor::ShiftBackspace()
{
   if(readOnly) return;

   // Steht der Cursor nicht auf Position 1: Lschen bis Position 1
   if(cursorX > 0) {
      char row[maxCols + 4];

      edLine.GetText(row);
      if(cursorX > (int)strlen(row)) {
         row[cursorX] = 0x00;
      }
      edLine.SetText(&row[cursorX]);
      SaveCurrentLine();
      SetCursor(cursorY,0);
      changed = true;
   }
   // Sonst anfgen an vorherige Zeile
   else {
      Backspace();
   }
}



// #######################################################################                                                                    
// ===== Cursor-Bewegungen ===============================================
// #######################################################################                                                                    


/* Cursor nach links bewegen */
void Editor::CursorLeft()
{
   if(cursorX>0) {
      edLine.CursorLeft();
      RethinkXPosition();
   }
   else {
      if(cursorY>0) {
         CursorUp();
         CursorEnd();
      }
   }
}


/* Cursor nach rechts bewegen */
void Editor::CursorRight()
{
   if( (cursorX < lineMgr.GetLineLength(cursorY)) || (!(autoCursorJump)) ) {
      edLine.CursorRight();
      RethinkXPosition();
   }
   else {
      if(cursorY < lineMgr.Count()) {
         CursorDown();   
         CursorHome();
      }
   }
}


/* Cursor nach oben bewegen */
void Editor::CursorUp()
{
   int l;

   if(cursorY > 0) {
      offsetX = edLine.Offset();
      cursorX = edLine.Cursor();
      l = lineMgr.GetLineLength(cursorY-1);
    
      if(cursorY == offsetY) {
         if(fastScrollY == 0) {
            offsetY--;
            cursorY--;
            UpdateDisplay();
         }
         else {
            ScrollUp(1,height / fastScrollY);    
         }
      }
      else {
         cursorY--;
         UpdateEditLine();
      }
  
      if( (cursorX >= l) && (autoCursorJump) ) {
         CursorHome();
      }
   }
   else {
      status = topOfText;
   }
}


/* Cursor nach unten */
void Editor::CursorDown()
{
   int l;

   if((cursorY < maxRows - 1) && (cursorY < lineMgr.Count())) {
      offsetX = edLine.Offset();
      cursorX = edLine.Cursor();
      l = lineMgr.GetLineLength(cursorY + 1);
     
      if(cursorY >= offsetY + height - 1) {
         if(fastScrollY == 0) {
            offsetY++;
            cursorY++;
            UpdateDisplay();
         }
         else {
            ScrollDown(1,height / fastScrollY);
         }
      }
      else {
         cursorY++;
         UpdateEditLine();
      }
   
      if( (cursorX >= l) && (autoCursorJump) ) {
        CursorEnd();
      } 
   }    
   else {
      status = endOfText;
   }
}


/* Aufwrts scrollen */
void Editor::ScrollUp(int subCursor, int subOffset)
{
   if(cursorY > 0) {
      cursorY -= subCursor;
      offsetY -= subOffset;
      if(cursorY < 0) cursorY = 0;
      if(offsetY < 0) offsetY = 0;

      UpdateDisplay();
     }
    else {
       status = topOfText;
    }
}


/* Abwrts scrollen */
void Editor::ScrollDown(int addCursor, int addOffset)
{
   if((cursorY < maxRows - 1) && (cursorY < lineMgr.Count())) {
      cursorY += addCursor;
      offsetY += addOffset;
      if(cursorY > lineMgr.Count()) {
         cursorY = lineMgr.Count();
      }
      if(offsetY > lineMgr.Count() - height) {
         offsetY = lineMgr.Count() - height + 2;
        if(offsetY < 0) offsetY = 0;   
      }
    
      UpdateDisplay();
   }
   else {
      status = endOfText;
   }
}


/* Cursor positionieren */
void Editor::SetCursor(int y, int x)
{
   int oox = offsetX;
   int ooy = offsetY;
   int ocx = cursorX;
   int ocy = cursorY;

   cursorX = x;
   cursorY = y;

   // X-Offset bei Bedarf neu berechnen
   if(!( ((x >= offsetX) && (x < offsetX + width)) )) {  
      // Offset ist Null, wenn x auf der ersten Seite ist       
      if(x < width) {
         offsetX = 0;
      }
      // Ansonsten Zentrierung auf Seitenmitte    
      else {
         offsetX = x - (width / 2);   
         if(offsetX < 0) {
            offsetX = 0;
         }
      }
   }

   // Y-Offset bei Bedarf neu berechnen
   if(!( ((y >= offsetY) && (y < offsetY + height)) )) {
      // Offset ist Null, wenn y auf der ersten Seite ist
      if(y < height) {
         offsetY = 0;
      }
      // Ansonsten Zentrierung auf Seitenmitte
      else {
         offsetY = y - (height / 2);
         if(offsetY < 0) {
            offsetY = 0;
         }
      }
   }

   // Bei Bedarf gesamtes Display updaten
   if(skipUpdate == 0) {
      if((oox != offsetX) || (ooy != offsetY)) {
         UpdateDisplay();
      }
      // Sonst nur teilweises Update
      else if((ocx != cursorX) || (ocy != cursorY)) {
         UpdateEditLine();
      }
   }
   else {
      UpdateEditLine();
   }
}


/* Zum Textanfang springen */
void Editor::JumpToTop()
{
   offsetX = 0; 
   offsetY = 0;
   cursorX = 0; 
   cursorY = 0;
   UpdateDisplay();
}


/* Zum Textende springen */
void Editor::JumpToEnd()
{
   int y = lineMgr.Count();
   SetCursor(y,lineMgr.GetLineLength(y));
}


// #######################################################################                                                                    
// ===== Datei-Operationen ===============================================
// #######################################################################                                                                    


/* Neuen Text beginnen */
void Editor::NewText()
{
   cursorX = 0; 
   offsetX = 0;
   cursorY = 0; 
   offsetY = 0;
   UnmarkBlock(false);
   lineMgr.DeleteAll();
   UpdateDisplay();
   changed = false;       
}


/* Text laden */
bool Editor::LoadText(char* name)
{
   ifstream is(name);
   return(LoadText(is));
}


/* Text laden */
bool Editor::LoadText(ifstream& is)
{
   if(is.good()) {
      int  line = 0;
      char row[maxCols + 4];
  
      NewText();
      while(!is.eof()) {
         is.getline(row,maxCols);
         int ok = lineMgr.InsertLine(line);
         if(ok == true) {
            lineMgr.SetLine(line,row);
         }
         else {
            status = bufferFull;
            break;
         }
         line++;
      }
      changed = false;
      UpdateDisplay();
      return(true);
   }
   return(false);
}


/* Text speichern */
bool Editor::SaveText(char* name)
{
   if(autoBackup) {
      char backupName[strlen(name) + 16];
      strcpy(backupName,name);
      strcat(backupName,"~");
      remove(backupName);
      rename(name,backupName);
   }
   ofstream os(name);
   return(SaveText(os));
}


/* Text speichern */
bool Editor::SaveText(ofstream& os)
{
   if(os.good()) {
      int  lines = lineMgr.Count();
      char row[maxCols + 4];

      for(int i = 0;i <= lines;i++) {
         lineMgr.GetLine(i,row);
         os << row;
         if(i < lines - 1) {
            os << endl;
         }
      }
      changed = false;
      return(true);
   }
   return(false);
}


// #######################################################################                                                                    
// ===== Einstellungen ndern ============================================
// #######################################################################                                                                    


/* Tabulatortaste */
void Editor::Tab(int i)
{
   int j;

   for(j=0;j<i;j++) {
      if(InsertChar(' ')==false) break;
      if(((cursorX+1) % i)==0) break;
   }
}


// #######################################################################
// ===== Blockoperationen ================================================
// #######################################################################


/* Markierung beginnen */
void Editor::BlockStart()
{
   UnmarkBlock(true);
   blockStartX = cursorX;
   blockStartY = cursorY;
   blockStatus = blockStart;
}


/* Markierung beenden */
#define Tausche(a,b) { i=a; a=b; b=i; }
void Editor::BlockEnd()
{
   int i;
   int ox1 = markX1;
   int oy1 = markY1;
   int ox2 = markX2;
   int oy2 = markY2;
   int ost = blockStatus;

   // Alte Markierung lschen
   UnmarkBlock(false);

   // Neue Markierung seten
   blockEndX = cursorX;
   blockEndY = cursorY;

   markX1 = blockStartX;
   markY1 = blockStartY;
   markX2 = blockEndX;
   markY2 = blockEndY;

   // Werte bei Bedarf tauschen
   if((markX1 == markX2) && (markY1 == markY2)) {
      blockStatus = noBlock;
   }
   else {
     blockStatus = blockEnd;
     if(markY1 > markY2) {
        Tausche(markY1,markY2);     
        Tausche(markX1,markX2);
     }
     else {
        if(markY1 == markY2) {
           if(markX1>markX2) {
              Tausche(markX1,markX2);
            }
         }
      }
   }
 
   if(ost == blockEnd) {        
      if((ox1 != markX1) || (oy1 != markY1)) {
         RefreshDisplay(min(oy1,markY1),max(oy1,markY1));
      }
      if((ox2 != markX2) || (oy2 != markY2)) {
         RefreshDisplay(min(oy2,markY2),max(oy2,markY2));      
      }
   }
   else {
      RefreshDisplay(markY1,markY2);
   }
}


/* Markierung aufhaben */
void Editor::UnmarkBlock(int update)
{
   if(!(blockStatus == noBlock)) {
      int start = markY1;
      int end   = markY2;
      markX1 = -1;
      markY1 = -1;          
      markX2 = -1;
      markY2 = -1;          
      blockStatus = noBlock;   

      if(update == true) {
         RefreshDisplay(start,end);
      }
   }
}


/* Block ausschneiden, falls dies notwendig ist */
int Editor::OverwriteBlock()
{
   if((overwriteBlocks) && (blockStatus == blockEnd) ) {
      if( ((markX1 <= cursorX) && (markY1 <= cursorY)) &&    
          ((markX2 >  cursorX) && (markY2 >= cursorY)) ) {
         CutBlock(tmpName);
         return(true);
      }
   }
   return(false);
}


/* Markierung mit Cursor-Tasten: 
   Falls notwendig, neue Markierung beginnen, ansonsten alte fortseten */
void Editor::CheckBlock(void)
{
   if(blockStatus == noBlock) {
      BlockStart();
   }
   else if(blockStatus == blockEnd) {
      if(!( ((cursorX == markX1) && (cursorY == markY1)) )) {
         if(!( ((cursorX == markX2) && (cursorY == markY2)) )) {
            BlockStart();
         }
      }
   }
}


/* Block kopieren */
bool Editor::CopyBlock(char* name)
{
   ofstream os(name);
   return(CopyBlock(os));
}


/* Block kopieren */
bool Editor::CopyBlock(ofstream& os)
{
   if(blockStatus == blockEnd) {
      if(os.good()) {
         char row[maxCols + 4];

         // Zeilen von Y1 bis Y2 durchlaufen
         for(int i = markY1;i <= markY2;i++) {
            // Zeile aus dem Speicher lesen
            lineMgr.GetLine(i,row);

            // Falls die Markierung nur aus einer einzigen Zeile
            // besteht: Ausgabe des entsprechenden Bereiches
            if((i == markY1) && (i == markY2)) {
               row[markX2] = 0x00;
               if((int)strlen(row) >= markX1)     
                  os << &row[markX1];
               else {
                  for(int j = 0;j < markX1;j++) {
                     os << " ";
                  }
               }
            }
            // Ist i die erste Zeile, Zeile ab Position X1 ausgeben
            else if(i == markY1) {
               if((int)strlen(row) >= markX1) 
                  os << &row[markX1] << endl;
               else {
                  for(int j = 0;j < markX1;j++)
                     os << " ";
                  os << endl;                 
               }
            }
            // Ist i die lette Zeile, Zeile bis Position X2-1 ausgeben
            else if(i == markY2) {
               row[markX2] = 0x00; 
               os << row;
            }
            // Sonst gesamte Zeile ausgeben
            else {
               os << row;
               if(i < markY2) {
                  os << endl;
               }
            }
         }
      }
      else {
         return(false);
      }
   }
   return(true);
}


/* Block lschen */
void Editor::KillBlock()
{
   if((!readOnly) && (blockStatus == blockEnd)) {
      int i,j,l;
      changed = true;
  
      // Zeilen von Y1 bis Y2 durchlaufen
      for(i = markY1;i <= markY2;i++) {
         char row1[maxCols + 4];
         char row2[maxCols + 4];

         // Besteht die Markierung nur aus einer Zeile:
         // Lschen des entsprechenden Abschnittes
         if((i == markY1) && (i == markY2)) {
            lineMgr.GetLine(markY1,row1);
            for(j = markX1;j < markX2;j++) {
               Delete(row1,markX1);
            }
            lineMgr.SetLine(markY1,row1);
         }
         // Ist i die erste Zeile, Lschen bis um Ende
         else if(i == markY1) {
            lineMgr.GetLine(markY1,row1);
            l = strlen(row1);
            for(j = markX1;j < l;j++) { 
               Delete(row1,markX1);
            }
         }
         // Falls i letzte Zeile ist:
         else if(i == markY2) {
            // Zeile bis X2 lschen
            lineMgr.GetLine(markY1 + 1,row2);
            for(j = 0;j < markX2;j++) { 
               Delete(row2,0); 
            }
  
            // Anhngen des Rests an die erste Zeile
            j = strlen(row1);
            l = strlen(row2);    
            if(j + l < maxCols) {
               strcat(row1,row2); 
               lineMgr.SetLine(markY1,row1);
               lineMgr.DeleteLine(markY1 + 1);
            }
            else {
               lineMgr.SetLine(markY1,row1);
               lineMgr.SetLine(markY1 + 1,row2);     
               status = lineTooLong;
            }
         }
         // Sonst gesamte Zeile lschen
         else {
             lineMgr.DeleteLine(markY1);
         }
      }
    
      // Markierung aufheben
      i = markX1;                     
      j = markY1;
      UnmarkBlock(false);

      // Update des Displays
      skipUpdate++;
      SetCursor(j,i);
      skipUpdate--;
      UpdateDisplay();
   }
}


/* Block ausschneiden */
bool Editor::CutBlock(char* name)
{
   bool result = CopyBlock(name);
   KillBlock();
   return(result);
}


/* Block ausschneiden */
bool Editor::CutBlock(ofstream& os)
{
   bool result = CopyBlock(os);
   KillBlock();
   return(result);
}


/* Block einfgen */
bool Editor::InsertBlock(char* name)
{
   ifstream is(name);
   return(InsertBlock(is));
}


/* Block einfgen */
bool Editor::InsertBlock(istream& is)
{
   if((!readOnly) && (is.good())) {
      int  i,j;
      char row0[maxCols + 4];
      char row1[maxCols + 4];
      char row2[maxCols + 4];
      char row3[maxCols + 4];
      
      // Evtl. vorhandene Markierung lschen
      UnmarkBlock(false);
      changed = true; 

      // Neue Markierung fr eingefgten Block
      markX1 = cursorX;
      markY1 = cursorY;

      // Aktuelle Zeile teilen 
      lineMgr.GetLine(cursorY,row1);
      Copy(row1,0,cursorX - 1,row2);
      Copy(row1,cursorX,strlen(row1) - 1,row3);

      i = cursorY;
      while(!is.eof()) {
         // Zeile einlesen
         is.getline(row1,maxCols);

         // Neue Zeile einfgen
         j = lineMgr.InsertLine(i);
         if(j == true) {

            // Zeilenende durch Null erseten:
            // Zeilenende kann CR oder CR + LF sein.
            j = strlen(row1);
            if(row1[j] == '\n') row1[j] = 0x00;
            j--;
            if(j >= 0) {
               if(row1[j] == '\n') {
                  row1[j] = 0x00;
               }
            }
 
            // Zeile in Zeilenspeicher kopieren
            lineMgr.SetLine(i,row1);
            i++;
         }
         // Fehler: Zeilenspeicher ist voll
         else {
            status = bufferFull;
            break;
         }
      }
   
      if(i > cursorY) {
         // Erste Zeile vervollstndigen
         lineMgr.GetLine(cursorY,row1);
         strcpy(row0,row2);
         if((int)strlen(row2) + (int)strlen(row1) < maxCols)
           strcat(row0,row1);
         else
            status = lineTooLong;
         lineMgr.SetLine(cursorY,row0);
     
         // Lette Zeile vervollstndigen
         lineMgr.GetLine(i - 1,row1);
         strcpy(row0,row1);
         if((int)strlen(row1) + (int)strlen(row3) < maxCols)
            strcat(row0,row3);
         else
            status = lineTooLong;
         lineMgr.SetLine(i - 1,row0);
         lineMgr.DeleteLine(i);   
      } 

      // Markierung seten
      markX2      = strlen(row1);
      markY2      = i - 1;
      blockStatus = blockEnd;

      // Cursor neu seten
      skipUpdate++;
      SetCursor(i - 1,strlen(row1));
      skipUpdate--;
      UpdateDisplay();
      return(true);
   }
   else {
      return(false);
   }
}
