// ##########################################################################
// ####                                                                  ####
// ####      WarpEd - A portable text editor for OS/2 Warp and Linux     ####
// ####    ===========================================================   ####
// ####                                                                  ####
// #### os2.cc                                                           ####
// ####                                                                  ####
// #### Version 0.80  --  October 06, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 1996  Thomas Dreibholz                             ####
// ####               1997  Molbachweg 7                                 ####
// ####                     51674 Wiehl/Germany                          ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>
#include <strings.h>       

#include "os2.h"

#include <signal.h>


// #######################################################################
// ===== Initialisierung =================================================
// #######################################################################


/* Ctrl-C-Handler fr OS/2 */
int zControlCDetected;
void ControlCHandler(int dummy)
{
   zControlCDetected = 1;
   signal(SIGINT,SIG_ACK);
}


/* Terminal initialisieren */
int OutputTerminalClass::OpenTerminal(ColorPrefs *cprefs)
{
   memcpy(&zColorPrefs,cprefs,sizeof(ColorPrefs));
   
   VIOMODEINFO vinfo;
   VioGetMode((PVIOMODEINFO)&vinfo,0);
   zWidth  = vinfo.col;
   zHeight = vinfo.row;

   int w = zWidth + zWidth; 
   int i = DosAllocMem((void **)&zVideoBackup,
                       w * zHeight,
                       PAG_READ|PAG_WRITE|PAG_COMMIT);
   if(i != 0) { 
      return(FALSE);
   }
 
   int j = 0;            
   for(i = 0;i < zHeight;i++) {
      VioReadCellStr((unsigned char *)&zVideoBackup[j],
                     (PUSHORT)&w,i,0,0);
      j += w;
   }
   VioGetCurPos((PUSHORT)&zOldCursorY,(PUSHORT)&zOldCursorX,0);        
   VioSetAnsi(1,0);
   zControlCDetected = 0;
   signal(SIGINT,&ControlCHandler);

   zWriteLevel = 0;
   zCursorX    = 0;
   zCursorY    = 0;
   zShowCursor = FALSE;
   Clear();
   return(TRUE);
}


/* Terminal schlieen */
void OutputTerminalClass::CloseTerminal()
{
   Clear();

   if(zVideoBackup != NULL) {
      int w = zWidth + zWidth;
      int j = 0;
      for(int i = 0;i < zHeight;i++) {
         VioWrtCellStr((unsigned char *)&zVideoBackup[j],w,i,0,0);
         j += w; 
      }
      DosFreeMem(zVideoBackup);
      VioSetCurPos(zOldCursorY,zOldCursorX,0);
   }
   signal(SIGINT,SIG_IGN);
}


// #######################################################################
// ===== Ausgabe =========================================================
// #######################################################################



// Ausgabe beginnen
void OutputTerminalClass::BeginWriting()
{ 
   zWriteLevel++;
   if(zWriteLevel == 1) {
      VioSetCurPos(26,81,0);
   }
}


// Ausgabe beenden
void OutputTerminalClass::EndWriting()
{ 
   zWriteLevel--;
   if(zWriteLevel == 0) {    
      VioSetCurPos(zCursorY,zCursorX,0);
   }
}
   

/* String-Ausgabe */
void OutputTerminalClass::WriteString(int x, int y, int style,
                                int frontPen, int backPen, char *str)
{
   char  v[2]; 
   int   l = strlen(str);
   VioWrtCharStr((unsigned char *)str,l,y,x,0);
   v[0] = (char)(frontPen + (16 * backPen));
   VioWrtNAttr((unsigned char *)v,l,y,x,0);
}


/* 
Zeile ausgeben
x,y:       Koordinaten
width:     Breite der Zeile
offset:    X-Offset
m1,m2:     Markierung
buffer:    Zeiger auf Buffer
*/
void OutputTerminalClass::WriteLine(int x, int y, int width, int offset, 
                              int m1, int m2, 
                              char       *buffer)
{
   int  i,j,k;
   int  len;
   int  frontPen;
   int  backPen;
   char ctrl[32];  
   char out[256];

   // Keine Markierung, wenn beide Positionen gleich sind
   if(m1 == m2) {
      m1 = -1;        
      m2 = -1;
   }

   // Ausgabe auf Offset beginnen: Anfang abschneiden bzw. leerer String
   len = strlen(buffer);
   if(len < offset) {
      ctrl[15] = 0x00;
      buffer   = &ctrl[15];
      len      = 0;
   }
   else {
      buffer += offset;
      len    -= offset;
   }
   
   // Markierung beginnt direkt am Anfang
   if((m1 != -1) && (m1 < offset)) { 
      frontPen = zColorPrefs.BlockFrontPen; 
      backPen  = zColorPrefs.BlockBackPen;
   } 
   else {
      frontPen = zColorPrefs.FrontPen;
      backPen  = zColorPrefs.BackPen;
   }

   int fp = frontPen;
   int bp = backPen;
   for(i = 0,j = 0,k = offset;i  < width;i++,j += 2,k++) {
      if(k == m1) {
         fp = zColorPrefs.BlockFrontPen;
         bp = zColorPrefs.BlockBackPen;
      }
      else if(k == m2) {
         fp = zColorPrefs.FrontPen;
         bp = zColorPrefs.BackPen;
      }
      out[j + 1] = fp + (16 * bp);
      if(i < len)
         out[j] = buffer[i];
      else
         out[j] = ' ';
   }
   VioWrtCellStr((unsigned char *)&out,j,y,x,0);
}


/* Bildschirm lschen */
void OutputTerminalClass::Clear()
{
   char v[2];
   v[0] = ' ';
   v[1] = 0x00;
   VioWrtNCell((unsigned char *)v,zWidth * zHeight,0,0,0);
}


// Sound-Ausgabe
void OutputTerminalClass::Bell()
{ 
}


// #######################################################################
// ===== Cursor-Verwaltung ===============================================
// #######################################################################


// Cursor setzen
void OutputTerminalClass::ShowSysCursor(int x, int y)
{ 
   zCursorX    = x;
   zCursorY    = y;
   zShowCursor = TRUE;
   if(zWriteLevel == 0) {
      VioSetCurPos(zCursorY,zCursorX,0);
   }
}  


// Cursor ausschalten
void OutputTerminalClass::HideSysCursor()
{
   zShowCursor = FALSE;
   if(zWriteLevel == 0) {
      VioSetCurPos(26,81,0);
   }
}


// #######################################################################
// ===== Eingabe =========================================================
// #######################################################################


/* Zeichen lesen */
int OutputTerminalClass::WaitForKey(int secs, int usecs)
{
   int key;
   static KBDKEYINFO ki;
   short             flag;
   unsigned char     chr,scan,status;
 
   if((secs != 0) || (usecs != 0)) { 
      flag = IO_WAIT; 
   }
   else {
      flag = IO_NOWAIT;
   }

   chr = scan = status = 0x00;
   key = 0;
   KbdCharIn(&ki,flag,0);

   if(zControlCDetected) {
      zControlCDetected = 0;
      key = ControlSeq('C');
   }
   else {
      chr    = ki.chChar;
      scan   = ki.chScan;
      status = ki.fsState;
      
      if((chr == 0x00) || (chr == 0xe0)) {
         key = chr + (scan << 8);
         if((status & KBDSTF_LEFTSHIFT) || (status & KBDSTF_RIGHTSHIFT))
            key |= KeyShift;
      }   
      else {
         key = chr;
  
         if(status & KBDSTF_CONTROL) {
            key |= KeyControl;
            key += 64;
         }
         else if(status & KBDSTF_ALT) {
            key |= KeyAlt;
         }
         if((chr == KeyBackspace) && 
           ((status & KBDSTF_LEFTSHIFT) || (status & KBDSTF_RIGHTSHIFT)) ) {
              key = KeyShiftBackspace;
         }
      }
   }
   return(key);
}
